/* based off: https://lkml.org/lkml/2007/7/11/100 */
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>

/*
 * Range of signals to send.  Each signal is sent 3 times.
 */
#define SIGSTART	SIGRTMIN
#define SIGEND		SIGSTART + 3
#define NUMSIG		((SIGEND - SIGSTART) * 3)

/*
 * The `seq` struct and `sh` handler are used to record the order in which
 * the signals are delivered to the parent process.  When a signal is
 * delivered the handler records it in seq and then increments the index.
 */
struct seq {
	int sig, num;
};
static struct seq *seqp;

void sh(int sig, siginfo_t *info, void *unused)
{
	static int cnt;
	seqp[cnt].num = info->si_value.sival_int;
	seqp[cnt++].sig = sig;
}

int main(int argc, char *argv[])
{
	int i, j, x, mask = 0, no_defer = 0;
	pid_t p = getpid(), chld;
	struct sigaction sa;
	sigset_t ss, ssold;
	union sigval sv;

	seqp = malloc(sizeof (struct seq) * NUMSIG);

	while((x = getopt(argc, argv, "m(mask)n(no-defer)")) != -1) {
		switch (x) {
		case 'm':
			mask = 1;
			break;
		case 'n':
			no_defer = 1;
			break;
		case -1:
			break;
		default:
			fprintf(stderr, "USAGE: ./rtsignal [--mask|-m] [--no-defer|-n]\n");
			exit(1);

		}
	}

	/*
	 * step 1:
	 *
	 * Install the signal handler (sh) on this PID, the parent.  If
	 * `mask` is passed as argument the block _all_ RT signals while
	 * running the handler.  If `dont` is passed then don't block any
	 * signals on the handler.
	 */
	sigemptyset(&ss);
	for (i = SIGSTART; i < SIGEND; i++) {
		sa.sa_sigaction = sh;
		sa.sa_flags = SA_SIGINFO;
		sa.sa_flags |= (no_defer ? SA_NODEFER : 0);
		if (mask) {
			sigfillset(&sa.sa_mask);
		} else {
			sigemptyset(&sa.sa_mask);
		}
		sigaction(i, &sa, NULL);
		sigaddset(&ss, i);
	}

	/* step 2:
	 *
	 * Block all signals that are about to be sent to this process.
	 * This is needed to make sure all signals are queued before they
	 * are delivered to the handler.  This is related to step 4.
	 */
 	sigprocmask(SIG_BLOCK, &ss, &ssold);

	/*
	 * step 3:
	 *
	 * Fork a child process to send each RT signal 3 times to the
	 * parent process.  Remember, all signals being sent are RT
	 * signals and thus are subject to explicit ordering guarantees
	 * spelled out in POSIX.1b.  Lower numbered RT signals are
	 * _always_ delivered before higher RT signals.
	 */
	if ((chld = fork()) == 0) {
		for (i = SIGSTART; i < SIGEND; i++) {
			for (j = 0; j <= 2; j++) {
				sv.sival_int = j;
				sigqueue(p, i, sv);
			}
		}
		return 0;
	}

	/* step 4:
	 *
	 * Wait for forked process to finish sending _all_ signals.  This
	 * is to make sure that all signals are queued up before allowing
	 * the kernel to deliver them to the handler.  If you don't both
	 * wait and block then the handler will process the signals as
	 * they come in and thus we will not be testing the ordering
	 * guarentees.  Related to step 2.
	 */
	while (wait(NULL) != chld) {
		;;
	}
	sigprocmask(SIG_SETMASK, &ssold, NULL);

	/* step 5:
	 *
	 * Print the order in which the signals were delivered to this
	 * process as recorded in the `seq` structure by the `sh` handler.
	 */
	for (i = 0; i < NUMSIG; i++) {
		printf("RT sig %2d received with sival %d\n",
		    seqp[i].sig, seqp[i].num);
	}

	free(seqp);
	return 0;
}
